@@ -9,10 +9,10 @@ from decimal import Decimal
 import datetime
 import time
 import io
-import collections
+from collections import OrderedDict
 
 from tryton.translate import date_format
-from tryton.common import untimezoned_date, datetime_strftime
+from tryton.common import untimezoned_date, timezoned_date, datetime_strftime
 from tryton.pyson import PYSONDecoder
 
 __all__ = ['DomainParser']
@@ -117,12 +117,20 @@ def quote(value):
     "Quote string if needed"
     if not isinstance(value, basestring):
         return value
+    if '"' in value:
+        value = value.replace('"', '\\"')
     for test in (':', ' ', '(', ')') + OPERATORS:
         if test in value:
             return '"%s"' % value
     return value
 
 
+def test_quote():
+    assert quote('test') == 'test'
+    assert quote('foo bar') == '"foo bar"'
+    assert quote('"foo"') == '\\\"foo\\\"'
+
+
 def ending_clause(domain, deep=0):
     "Return the ending clause"
     if not domain:
@@ -159,7 +167,8 @@ def append_ending_clause(domain, clause, deep):
 
 def default_operator(field):
     "Return default operator for field"
-    if field['type'] in ('char', 'text', 'many2one', 'many2many', 'one2many'):
+    if field['type'] in ('char', 'text', 'many2one', 'many2many', 'one2many',
+            'reference'):
         return 'ilike'
     else:
         return '='
@@ -179,12 +188,44 @@ def time_format(field):
     return PYSONDecoder({}).decode(field['format'])
 
 
+def split_target_value(field, value):
+    "Split the reference value into target and value"
+    assert field['type'] == 'reference'
+    target = None
+    if isinstance(value, basestring):
+        for key, text in field['selection']:
+            if value.lower().startswith(text.lower() + ','):
+                target = key
+                value = value[len(text) + 1:]
+                break
+    return target, value
+
+
+def test_split_target_value():
+    field = {
+        'type': 'reference',
+        'selection': [
+            ('spam', 'Spam'),
+            ('ham', 'Ham'),
+            ('e', 'Eggs'),
+            ]
+        }
+    for value, result in (
+            ('Spam', (None, 'Spam')),
+            ('foo', (None, 'foo')),
+            ('Spam,', ('spam', '')),
+            ('Ham,bar', ('ham', 'bar')),
+            ('Eggs,foo', ('e', 'foo')),
+            ):
+        assert split_target_value(field, value) == result
+
+
 def convert_value(field, value):
     "Convert value for field"
 
     def convert_boolean():
         if isinstance(value, basestring):
-            return any(test.lower().startswith(value.lower())
+            return any(test.decode('utf-8').lower().startswith(value.lower())
                 for test in (_('y'), _('yes'), _('true'), _('t'), '1'))
         else:
             return bool(value)
@@ -328,13 +369,18 @@ def test_convert_selection():
             ('female', 'Female'),
             ],
         }
+    field_with_empty = field.copy()
+    field_with_empty['selection'] = (field_with_empty['selection']
+        + [('', '')])
     for value, result in (
             ('Male', 'male'),
             ('male', 'male'),
             ('test', 'test'),
             (None, None),
+            ('', ''),
             ):
         assert convert_value(field, value) == result
+        assert convert_value(field_with_empty, value) == result
 
 
 def test_convert_datetime():
@@ -344,7 +390,8 @@ def test_convert_datetime():
         }
     for value, result in (
             ('12/04/2002', datetime.datetime(2002, 12, 4)),
-            ('12/04/2002 12:30:00', datetime.datetime(2002, 12, 4, 12, 30)),
+            ('12/04/2002 12:30:00', untimezoned_date(
+                    datetime.datetime(2002, 12, 4, 12, 30))),
             ('test', None),
             (None, None),
             ):
@@ -376,7 +423,7 @@ def test_convert_time():
         assert convert_value(field, value) == result
 
 
-def format_value(field, value):
+def format_value(field, value, target=None):
     "Format value for field"
 
     def format_boolean():
@@ -393,15 +440,21 @@ def format_value(field, value):
                 and not isinstance(value, (float, Decimal))):
             return ''
         try:
-            digit = int(field.get('digits', (16, 2))[1])
-        except ValueError:
-            digit = 2
-        return locale.format('%.*f', (digit, value or 0.0), True)
+            digit = len(str(value).split('.')[1])
+        except IndexError:
+            digit = 0
+        return locale.format('%.*f', (digit, value or 0), True)
 
     def format_selection():
         selections = dict(field['selection'])
         return selections.get(value, value) or ''
 
+    def format_reference():
+        if not target:
+            return format_selection()
+        selections = dict(field['selection'])
+        return '%s,%s' % (selections.get(target, target), value)
+
     def format_datetime():
         if not value:
             return ''
@@ -409,7 +462,10 @@ def format_value(field, value):
         if (not isinstance(value, datetime.datetime)
                 or value.time() == datetime.time.min):
             format_ = date_format()
-        return datetime_strftime(value, format_)
+            time = value
+        else:
+            time = timezoned_date(value)
+        return datetime_strftime(time, format_)
 
     def format_date():
         if not value:
@@ -432,7 +488,7 @@ def format_value(field, value):
         'float': format_float,
         'numeric': format_float,
         'selection': format_selection,
-        'reference': format_selection,
+        'reference': format_reference,
         'datetime': format_datetime,
         'date': format_date,
         'time': format_time,
@@ -474,13 +530,14 @@ def test_format_integer():
 def test_format_float():
     field = {
         'type': 'float',
-        'digits': (16, 2),
         }
     for value, result in (
-            (1, '1.00'),
-            (1.5, '1.50'),
-            (0, '0.00'),
-            (0.0, '0.00'),
+            (1, '1'),
+            (1.5, '1.5'),
+            (1.50, '1.5'),
+            (150.79, '150.79'),
+            (0, '0'),
+            (0.0, '0.0'),
             (False, ''),
             (None, ''),
             ):
@@ -490,13 +547,14 @@ def test_format_float():
 def test_format_numeric():
     field = {
         'type': 'numeric',
-        'digits': (16, 2),
         }
     for value, result in (
-            (Decimal(1), '1.00'),
-            (Decimal('1.5'), '1.50'),
-            (Decimal(0), '0.00'),
-            (Decimal('0.0'), '0.00'),
+            (Decimal(1), '1'),
+            (Decimal('1.5'), '1.5'),
+            (Decimal('1.50'), '1.50'),
+            (Decimal('150.79'), '150.79'),
+            (Decimal(0), '0'),
+            (Decimal('0.0'), '0.0'),
             (False, ''),
             (None, ''),
             ):
@@ -511,13 +569,18 @@ def test_format_selection():
             ('female', 'Female'),
             ],
         }
+    field_with_empty = field.copy()
+    field_with_empty['selection'] = (field_with_empty['selection']
+        + [('', '')])
     for value, result in (
             ('male', 'Male'),
             ('test', 'test'),
             (False, ''),
             (None, ''),
+            ('', ''),
             ):
         assert format_value(field, value) == result
+        assert format_value(field_with_empty, value) == result
 
 
 def test_format_datetime():
@@ -528,7 +591,8 @@ def test_format_datetime():
     for value, result in (
             (datetime.date(2002, 12, 4), '12/04/2002'),
             (datetime.datetime(2002, 12, 4), '12/04/2002'),
-            (datetime.datetime(2002, 12, 4, 12, 30), '"12/04/2002 12:30:00"'),
+            (untimezoned_date(datetime.datetime(2002, 12, 4, 12, 30)),
+                '"12/04/2002 12:30:00"'),
             (False, ''),
             (None, ''),
             ):
@@ -573,6 +637,7 @@ def complete_value(field, value):
         test_value = value if value is not None else ''
         if isinstance(value, list):
             test_value = value[-1]
+        test_value = test_value.strip('%')
         for svalue, test in field['selection']:
             if test.lower().startswith(test_value.lower()):
                 if isinstance(value, list):
@@ -580,9 +645,21 @@ def complete_value(field, value):
                 else:
                     yield svalue
 
+    def complete_reference():
+        test_value = value if value is not None else ''
+        if isinstance(value, list):
+            test_value = value[-1]
+        test_value = test_value.strip('%')
+        for svalue, test in field['selection']:
+            if test.lower().startswith(test_value.lower()):
+                if isinstance(value, list):
+                    yield value[:-1] + [svalue]
+                else:
+                    yield likify(svalue)
+
     def complete_datetime():
         yield datetime.date.today()
-        yield datetime.datetime.now()
+        yield datetime.datetime.utcnow()
 
     def complete_date():
         yield datetime.date.today()
@@ -593,7 +670,7 @@ def complete_value(field, value):
     completes = {
         'boolean': complete_boolean,
         'selection': complete_selection,
-        'reference': complete_selection,
+        'reference': complete_reference,
         'datetime': complete_datetime,
         'date': complete_date,
         'time': complete_time,
@@ -618,6 +695,18 @@ def test_complete_selection():
             ):
         assert list(complete_value(field, value)) == result
 
+    field_with_empty = field.copy()
+    field_with_empty['selection'] = (field_with_empty['selection']
+        + [('', '')])
+    for value, result in (
+            ('m', ['male']),
+            ('test', []),
+            ('', ['male', 'female', '']),
+            (None, ['male', 'female', '']),
+            (['male', 'f'], [['male', 'female']]),
+            ):
+        assert list(complete_value(field_with_empty, value)) == result
+
 
 def parenthesize(tokens):
     "Nest tokens according to parenthesis"
@@ -719,11 +808,7 @@ class DomainParser(object):
     "A parser for domain"
 
     def __init__(self, fields):
-        if hasattr(collections, 'OrderedDict'):
-            odict = collections.OrderedDict
-        else:
-            odict = dict
-        self.fields = odict((name, f)
+        self.fields = OrderedDict((name, f)
             for name, f in fields.iteritems()
             if f.get('searchable', True))
         self.strings = dict((f['string'].lower(), f)
@@ -745,52 +830,78 @@ class DomainParser(object):
             if exception.message == 'No closing quotation':
                 return self.parse(input_ + '"')
 
+    def stringable(self, domain):
+
+        def stringable_(clause):
+            if not clause:
+                return True
+            if (((clause[0] in ('AND', 'OR'))
+                        or isinstance(clause[0], (list, tuple)))
+                    and all(isinstance(c, (list, tuple)) for c in clause[1:])):
+                return self.stringable(clause)
+            if clause[0] in self.fields or clause[0] == 'rec_name':
+                return True
+            return False
+
+        if not domain:
+            return True
+        if domain[0] in ('AND', 'OR'):
+            domain = domain[1:]
+        return all(stringable_(clause) for clause in domain)
+
     def string(self, domain):
         "Return string for the domain"
 
         def string_(clause):
             if not clause:
                 return ''
-            if (isinstance(clause[0], basestring)
-                    and (clause[0] in self.fields
-                    or clause[0] == 'rec_name')):
-                name, operator, value = clause
-                if name not in self.fields:
-                    escaped = value.replace('%%', '__')
-                    if escaped.startswith('%') and escaped.endswith('%'):
-                        value = value[1:-1]
-                    return quote(value)
-                field = self.fields[name]
-                if 'ilike' in operator:
-                    escaped = value.replace('%%', '__')
-                    if escaped.startswith('%') and escaped.endswith('%'):
-                        value = value[1:-1]
-                    elif '%' not in escaped:
-                        if operator == 'ilike':
-                            operator = '='
-                        else:
-                            operator = '!'
-                        value = value.replace('%%', '%')
-                def_operator = default_operator(field)
-                if (def_operator == operator.strip()
-                        or (def_operator in operator
-                            and 'not' in operator)):
-                    operator = operator.rstrip(def_operator
-                        ).replace('not', '!').strip()
-                if operator.endswith('in'):
-                    if operator == 'not in':
-                        operator = '!'
-                    else:
-                        operator = ''
-                formatted_value = format_value(field, value)
-                if (operator in OPERATORS and
-                        field['type'] in ('char', 'text', 'sha', 'selection')
-                        and value == ''):
-                    formatted_value = '""'
-                return '%s: %s%s' % (quote(field['string']), operator,
-                    formatted_value)
-            else:
+            if (not isinstance(clause[0], basestring)
+                    or clause[0] in ('AND', 'OR')):
                 return '(%s)' % self.string(clause)
+            name, operator, value = clause[:3]
+            if name.endswith('.rec_name'):
+                name = name[:-9]
+            if name not in self.fields:
+                escaped = value.replace('%%', '__')
+                if escaped.startswith('%') and escaped.endswith('%'):
+                    value = value[1:-1]
+                return quote(value)
+            field = self.fields[name]
+
+            if len(clause) > 3:
+                target = clause[3]
+            else:
+                target = None
+
+            if 'ilike' in operator:
+                escaped = value.replace('%%', '__')
+                if escaped.startswith('%') and escaped.endswith('%'):
+                    value = value[1:-1]
+                elif '%' not in escaped:
+                    if operator == 'ilike':
+                        operator = '='
+                    else:
+                        operator = '!'
+                    value = value.replace('%%', '%')
+            def_operator = default_operator(field)
+            if (def_operator == operator.strip()
+                    or (def_operator in operator
+                        and ('not' in operator
+                            or '!' in operator))):
+                operator = operator.rstrip(def_operator
+                    ).replace('not', '!').strip()
+            if operator.endswith('in'):
+                if operator == 'not in':
+                    operator = '!'
+                else:
+                    operator = ''
+            formatted_value = format_value(field, value, target)
+            if (operator in OPERATORS and
+                    field['type'] in ('char', 'text', 'selection')
+                    and value == ''):
+                formatted_value = '""'
+            return '%s: %s%s' % (quote(field['string']), operator,
+                formatted_value)
 
         if not domain:
             return ''
@@ -841,8 +952,13 @@ class DomainParser(object):
         "Return all completion for the clause"
         if len(clause) == 1:
             name, = clause
-        else:
+        elif len(clause) == 3:
             name, operator, value = clause
+        else:
+            name, operator, value, target = clause
+            if name.endswith('.rec_name'):
+                name = name[:-9]
+            value = target
         if name == 'rec_name':
             if operator == 'ilike':
                 escaped = value.replace('%%', '__')
@@ -953,6 +1069,14 @@ class DomainParser(object):
                 else:
                     name, operator, value = clause
                     field = self.strings[name.lower()]
+                    field_name = field['name']
+
+                    target = None
+                    if field['type'] == 'reference':
+                        target, value = split_target_value(field, value)
+                        if target:
+                            field_name += '.rec_name'
+
                     if operator is None:
                         operator = default_operator(field)
                     if isinstance(value, list):
@@ -962,8 +1086,6 @@ class DomainParser(object):
                             operator = 'in'
                     if operator == '!':
                         operator = negate_operator(default_operator(field))
-                    if 'like' in operator:
-                        value = likify(value)
                     if field['type'] in ('integer', 'float', 'numeric',
                             'datetime', 'date', 'time'):
                         if value and '..' in value:
@@ -971,15 +1093,40 @@ class DomainParser(object):
                             lvalue = convert_value(field, lvalue)
                             rvalue = convert_value(field, rvalue)
                             yield iter([
-                                    (field['name'], '>=', lvalue),
-                                    (field['name'], '<', rvalue),
+                                    (field_name, '>=', lvalue),
+                                    (field_name, '<', rvalue),
                                     ])
                             continue
                     if isinstance(value, list):
                         value = [convert_value(field, v) for v in value]
+                        if field['type'] in ('many2one', 'one2many',
+                                'many2many', 'one2one'):
+                            field_name += '.rec_name'
                     else:
                         value = convert_value(field, value)
-                    yield field['name'], operator, value
+                    if 'like' in operator:
+                        value = likify(value)
+                    if target:
+                        yield (field_name, operator, value, target)
+                    else:
+                        yield field_name, operator, value
+
+
+def test_stringable():
+    dom = DomainParser({
+            'name': {
+                'string': 'Name',
+                'type': 'char',
+                },
+            })
+    valid = ('name', '=', 'Doe')
+    invalid = ('surname', '=', 'John')
+    assert dom.stringable([valid])
+    assert not dom.stringable([invalid])
+    assert dom.stringable(['AND', valid])
+    assert not dom.stringable(['AND', valid, invalid])
+    assert dom.stringable([[valid]])
+    assert not dom.stringable([[valid], [invalid]])
 
 
 def test_string():
@@ -996,6 +1143,28 @@ def test_string():
                 'string': 'Date',
                 'type': 'date',
                 },
+            'selection': {
+                'string': 'Selection',
+                'type': 'selection',
+                'selection': [
+                    ('male', 'Male'),
+                    ('female', 'Female'),
+                    ('', ''),
+                    ],
+                },
+            'reference': {
+                'string': 'Reference',
+                'type': 'reference',
+                'selection': [
+                    ('spam', 'Spam'),
+                    ('ham', 'Ham'),
+                    ]
+                },
+            'many2one': {
+                'string': 'Many2One',
+                'name': 'many2one',
+                'type': 'many2one',
+                },
             })
     assert dom.string([('name', '=', 'Doe')]) == 'Name: =Doe'
     assert dom.string([('name', '=', None)]) == 'Name: ='
@@ -1028,6 +1197,20 @@ def test_string():
     assert dom.string([('surname', 'ilike', '%Doe%')]) == '"(Sur)Name": Doe'
     assert dom.string([('date', '>=', datetime.date(2012, 10, 24))]) == \
         'Date: >=10/24/2012'
+    assert dom.string([('selection', '=', '')]) == 'Selection: '
+    assert dom.string([('selection', '=', None)]) == 'Selection: '
+    assert dom.string([('selection', '!=', '')]) == 'Selection: !""'
+    assert dom.string([('selection', '=', 'male')]) == 'Selection: Male'
+    assert dom.string([('selection', '!=', 'male')]) == 'Selection: !Male'
+    assert dom.string([('reference', 'ilike', '%foo%')]) == \
+        'Reference: foo'
+    assert dom.string([('reference.rec_name', 'ilike', '%bar%', 'spam')]) == \
+        'Reference: Spam,bar'
+    assert dom.string([('reference', 'in', ['foo', 'bar'])]) == \
+        'Reference: foo;bar'
+    assert dom.string([('many2one', 'ilike', '%John%')]) == 'Many2One: John'
+    assert dom.string([('many2one.rec_name', 'in', ['John', 'Jane'])]) == \
+        'Many2One: John;Jane'
 
 
 def test_group():
@@ -1113,6 +1296,9 @@ def test_group():
         ('Name', '=', None),
         ('Name', None, 'Doe'),
         ]
+    assert rlist(dom.group(udlex(u'Name: \\"foo\\"'))) == [
+        ('Name', None, '"foo"'),
+        ]
 
 
 def test_parse_clause():
@@ -1136,6 +1322,20 @@ def test_parse_clause():
                     ('female', 'Female'),
                     ],
                 },
+            'reference': {
+                'string': 'Reference',
+                'name': 'reference',
+                'type': 'reference',
+                'selection': [
+                    ('spam', 'Spam'),
+                    ('ham', 'Ham'),
+                    ]
+                },
+            'many2one': {
+                'string': 'Many2One',
+                'name': 'many2one',
+                'type': 'many2one',
+                },
             })
     assert rlist(dom.parse_clause([('John',)])) == [
         ('rec_name', 'ilike', '%John%')]
@@ -1155,6 +1355,12 @@ def test_parse_clause():
     assert rlist(dom.parse_clause([('Name', '!', ['John', 'Jane'])])) == [
         ('name', 'not in', ['John', 'Jane']),
         ]
+    assert rlist(dom.parse_clause([('Selection', None, None)])) == [
+        ('selection', '=', None),
+        ]
+    assert rlist(dom.parse_clause([('Selection', None, '')])) == [
+        ('selection', '=', ''),
+        ]
     assert rlist(dom.parse_clause([('Selection', None, ['Male', 'Female'])])) \
         == [
             ('selection', 'in', ['male', 'female'])
@@ -1166,3 +1372,21 @@ def test_parse_clause():
             ('integer', '>=', 3),
             ('integer', '<', 5),
             ]]
+    assert rlist(dom.parse_clause([('Reference', None, 'foo')])) == [
+        ('reference', 'ilike', '%foo%'),
+        ]
+    assert rlist(dom.parse_clause([('Reference', None, 'Spam')])) == [
+        ('reference', 'ilike', '%spam%'),
+        ]
+    assert rlist(dom.parse_clause([('Reference', None, 'Spam,bar')])) == [
+        ('reference.rec_name', 'ilike', '%bar%', 'spam'),
+        ]
+    assert rlist(dom.parse_clause([('Reference', None, ['foo', 'bar'])])) == [
+        ('reference', 'in', ['foo', 'bar']),
+        ]
+    assert rlist(dom.parse_clause([('Many2One', None, 'John')])) == [
+        ('many2one', 'ilike', '%John%'),
+        ]
+    assert rlist(dom.parse_clause([('Many2One', None, ['John', 'Jane'])])) == [
+        ('many2one.rec_name', 'in', ['John', 'Jane']),
+        ]
