@@ -33,7 +33,7 @@ class Many2One(WidgetInterface):
         self.wid_text.connect('populate-popup', self._populate_popup)
         self.wid_text.connect('focus-out-event',
             lambda x, y: self._focus_out())
-        self.wid_text.connect_after('changed', self.sig_changed)
+        self.wid_text.connect('changed', self.sig_changed)
         manage_entry_position(self.wid_text)
         self.changed = True
         self.focus_out = True
@@ -308,7 +308,7 @@ class Many2One(WidgetInterface):
         if not self.changed:
             return False
         value = self.field.get(self.record)
-        if self.has_target(value):
+        if self.has_target(value) and self.modified:
             def clean():
                 if not self.wid_text.props.window:
                     return
@@ -316,9 +316,12 @@ class Many2One(WidgetInterface):
                 position = self.wid_text.get_position()
                 self.field.set_client(self.record,
                     self.value_from_id(None, ''))
-                # Restore text and position after display
-                self.wid_text.set_text(text)
-                self.wid_text.set_position(position)
+                # The value of the field could be different of None
+                # in such case, the original text should not be restored
+                if not self.wid_text.get_text():
+                    # Restore text and position after display
+                    self.wid_text.set_text(text)
+                    self.wid_text.set_position(position)
             gobject.idle_add(clean)
         return False
 
